# !/usr/bin/env python
# coding: utf-8
import os
import time
import datetime
from datetime import datetime
from pathlib import PurePath
from tqdm import tqdm

import geopy
import geopy.distance

import gpxpy.geo as gpxgeo

import numpy as np
import pandas as pd

from ssoss.static_road_object import Intersection, StaticRoadObject


class DynamicRoadObject:

    def __init__(self, id_num, name, obj_type, sro_df, gpx_df, source="GPX"):
        """Dynamic Road Objects move through time and space using a source (typ. GPX)

        :param id_num: unique ID number for object
        :param name: name for dynamic object
        :param obj_type: type of dynamic object (vehicle, ship, drone, etc.)
        :param sro_df: dataframe of static road objects
        :param gpx_df: dataframe of gpx points
        :param source: default to GPX, but in future could be a live stream of GPS data
        """

        self.MStoMPH = 2.23694
        self.FTPStoMPH = 0.681818
        self.MPHtoFTPS = 1/self.FTPStoMPH
        self.MStoFTPS = self.MStoMPH * self.MPHtoFTPS
        self.DATE_FORMAT = '%m-%d-%Y--%H-%M-%S.%f-%Z'

        # self.sorted_sroDF = None

        self.id_num = id_num
        self.name = name
        self.obj_type = obj_type
        self.gpx_df = gpx_df
        self.t0 = gpx_df.loc[0].t
        self.t1 = gpx_df.loc[1].t
        self.pnt0 = gpx_df.loc[0].geo_point
        self.pnt1 = gpx_df.loc[1].geo_point

        self.pt0 = geopy.Point(self.pnt0.latitude, self.pnt0.longitude)
        self.pt1 = geopy.Point(self.pnt1.latitude, self.pnt1.longitude)
        self.spd = gpx_df.loc[0].spd

        self.bearing = 0
        self.source = source

        self.sro_df = sro_df
        self.sorted_sroDF = None
        self.closest_intersection = self.get_closest_intersection(as_list=False)
        self.closest_intersection_list = self.get_closest_intersection(as_list=True)
        self.closest_approaching_intersection = self.get_closest_approaching_intersection(
        )

        self.in_file_path = PurePath("./in/")
        self.out_file_path = PurePath("./out/")

        if self.spd is None:
            self.calculate_spd_values()

    @staticmethod
    # https://stackoverflow.com/questions/11869910/pandas-filter-rows-of-dataframe-with-operator-chaining
    def mask(df, key, value) -> pd.DataFrame:
        return df[df[key] == value]

    def update_location_simple(self, i=2):
        """ Update dynamic object location with new data point i
        """
        self.t0 = self.t1
        self.t1 = self.gpx_df.loc[i].t

        self.pnt0 = self.pnt1
        self.pnt1 = self.gpx_df.loc[i].geo_point

        self.pt0 = self.pt1
        self.pt1 = geopy.Point(self.pnt1.latitude, self.pnt1.longitude)

        self.spd = self.gpx_df.loc[i].spd

        self.closest_intersection = self.get_closest_intersection()
        self.closest_approaching_intersection = self.get_closest_approaching_intersection()

    def first_timestamp(self) -> pd.Timestamp:
        t = datetime.fromisoformat(str(self.t0))
        return pd.Timestamp(t)

    def current_timestamp(self) -> pd.Timestamp:
        t = datetime.fromisoformat(str(self.t1))
        return pd.Timestamp(t)

    def first_utc_timestamp(self):
        t = self.t0.timetuple()
        return time.mktime(t) - 28800

    def get_prev_utc_timestamp(self) -> float:
        t = self.t0.timetuple()
        return time.mktime(t) - 28800

    def get_utc_timestamp(self) -> float:
        t = self.t1.timetuple()
        return time.mktime(t) - 28800

    @staticmethod
    def utc_to_timestamp(t):
        return time.asctime(time.localtime(t))

    def get_time_step(self):
        time_step = self.t1 - self.t0
        if time_step.total_seconds() < 0:
            return 10.0  # assume larger first gpx point time step
        else:
            return time_step.total_seconds()

    def get_location(self, i=None, elev=False):
        if elev:
            if i is None:
                return self.pnt1.format_decimal()
            else:
                self.update_location_simple(i)
                return self.pnt1.format_decimal()
        else:
            if i is None:
                return self.pt1.format_decimal()
            else:
                self.update_location_simple(i)
                return self.pt1.format_decimal()

    def get_dist_step(self) -> geopy.distance:
        """ first distance step

        :return: geopy distance in feet
        """
        return geopy.distance.distance(self.pt0, self.pt1).ft

    @staticmethod
    def get_dist_step_from_points(point1, point2) -> geopy.distance:
        """distance between two points in feet.

        :param point1: geopy point 1
        :param point2: geopy point 2
        :return: geopy distance in feet
        """

        return geopy.distance.distance(point1, point2).ft

    def cur_dist_to_sro(self, sro: StaticRoadObject) -> geopy.distance:
        if sro is None:
            return None
        else:
            return geopy.distance.distance(self.pt1, sro.get_location()).ft

    def prev_dist_to_sro(self, sro: StaticRoadObject) -> geopy.distance:
        if sro is None:
            return None
        else:
            return geopy.distance.distance(self.pt0, sro.get_location()).ft

    def get_spd(self, units="MPH") -> float:
        if units == "MPH":
            return float(self.spd * self.MStoMPH) # Ft/sec
        else:
            return float(self.spd)  # Meters/sec

    # Still necessary?
    def calculate_spd_values(self):
        """
        Loads speed values for all gpx_df items into a spd_list.

        This might be used if GPX v1.1 does not log speed data, so this will calculate it.
        """
        self.gpx_df.drop(['spd'], axis=1)  # remove None values for speed
        spd_list = [0]

        for n in range(1, self.gpx_df.last_valid_index()):
            self.update_location_simple(n)
            if self.get_time_step() == 0:
                speed = 0
            else:
                speed = (self.get_dist_step() /
                         self.get_time_step()) * self.FTPStoMPH

            spd_list.append(speed)
            if n == self.gpx_df.last_valid_index():
                self.gpx_df['spd'] = spd_list

    def get_bearing(self) -> float:
        b = gpxgeo.get_course(self.pt0[0], self.pt0[1], self.pt1[0], self.pt1[1])
        self.bearing = b
        return b

    def approaching(self, sro: StaticRoadObject) -> bool:  # , self.sro: StaticRoadObject
        if self.cur_dist_to_sro(sro) <= self.prev_dist_to_sro(sro):
            return True
        else:
            return False

    def get_closest_intersection(self, as_list=False) -> Intersection:
        """returns None or 1st or ascending sorted list of intersection objects based on distance
        """

        # Crop min and max distances to limit search, sort and length
        min_sd = self.sro_df.iloc[0, 1].get_sd("min")
        max_sd = self.sro_df.iloc[0, 1].get_sd("max")

        for row in range(0, self.sro_df.last_valid_index()):
            self.sro_df.loc[row,
                            "d"] = self.cur_dist_to_sro(self.sro_df.iloc[row, 1])
            self.sro_df.loc[row, "approaching"] = self.approaching(
                self.sro_df.iloc[row, 1])

        self.sorted_sroDF = self.sro_df.sort_values(by=['d'], ignore_index=True)

        min_row = self.sorted_sroDF[self.sorted_sroDF['d'].ge(min_sd)].index

        if len(min_row) == 0:
            min_row = 0
        else:
            min_row = min(min_row)

        max_row = self.sorted_sroDF[self.sorted_sroDF['d'].le(max_sd)].index
        if len(max_row) == 0:
            max_row = 0
        else:
            max_row = max(max_row) + 1  # +1 for inclusive

        limited_df = self.sorted_sroDF[min_row:max_row]

        if limited_df.empty:
            return None
        else:
            if as_list:
                return limited_df
            else:
                return limited_df.iloc[0, 1]

    def get_closest_approaching_intersection(self, as_list=False) -> Intersection:
        """ returns the closest, approaching intersection object for the current point
        of the dynamic object based on sorted list of intersections.

        :return: intersection object
        """
        df = self.get_closest_intersection(as_list=True)
        if df is None:
            return None
        else:
            mask = df['approaching'].values == True
            if df[mask].empty:
                return None
            elif as_list:
                return df[mask]
            elif as_list == False:
                return df[mask].iloc[0, 1]

    def calc_bearing_diff(self, m: float) -> float:
        n = self.get_bearing()
        b = min(abs(n - m), abs(360 - n + m), abs(360 - m + n))
        return b

    def approach_leg(self, itrsxn: Intersection, index_out=True):
        """determines the approach leg of the intersection based on compass headings

        :param itrsxn: intersection object to find approach leg of
        :param index_out: when True returns integer index, false returns string.
        :return: index of leg being approached of intersection based on bearing
             "True" index_out: 0 - North, 1 - East, 2 - South, 3 - West
             "False" index_out returns strings:  "NB","EB", "SB", "WB".
        """

        veh_int_diff = [
            self.calc_bearing_diff(itrsxn.get_bearing(0)),
            self.calc_bearing_diff(itrsxn.get_bearing(1)),
            self.calc_bearing_diff(itrsxn.get_bearing(2)),
            self.calc_bearing_diff(itrsxn.get_bearing(3))
        ]

        approach_leg_index = np.argmin(veh_int_diff)
        app_leg_dir = ""
        if itrsxn is None:
            return None
        elif index_out:
            return int(approach_leg_index)
        else:
            if approach_leg_index == 0:
                app_leg_dir = "NB"
            elif approach_leg_index == 1:
                app_leg_dir = "EB"
            elif approach_leg_index == 2:
                app_leg_dir = "SB"
            elif approach_leg_index == 3:
                app_leg_dir = "WB"
            return app_leg_dir

    def drive_gpx(self, gpx_filename: str, use_pickle_file=False) -> pd.DataFrame:
        """ Load the GPX file into a dataframe with timestamp, location, speed, dist to event, bearing, and
        id of approaching intersection

        :param gpx_filename: absolute filepath of file (without .gpx or .p file)
        :param use_pickle_file: default to False, can be faster to load from pickle
        :return: DataFrame with GPX file loaded and saves a CSV and Pickle file of gpx information
        """

        # gpx_file = PurePath(self.in_file_path, gpx_filename + ".gpx")
        pickle_file = self.out_file_path / (str(gpx_filename) + ".pkl")
        csv_file = self.out_file_path / (str(gpx_filename) + ".csv")

        approach_log_df = None

        if use_pickle_file and os.path.isfile(pickle_file):
            approach_log_df = pd.read_pickle(pickle_file)
            return approach_log_df
        else:
            # dictionary-> Keys:Values
            appr_dict = {
                "id": [],
                "appr_dir": [],
                "timestamp": [],
                "location": [],
                "spd": [],
                "distance": [],
                "bearing": [],
                "approaching": []
            }

            for i in range(2, self.gpx_df.last_valid_index()):
                self.update_location_simple(i)
                cai = self.get_closest_approaching_intersection()
                if cai is None:
                    if i == self.gpx_df.last_valid_index() - 1:
                        approach_log_df = pd.DataFrame(appr_dict)
                        approach_log_df.to_csv(csv_file)
                        approach_log_df.to_pickle(pickle_file)
                        print(
                            f"exported data frame to CSV (in {csv_file}) and Pickle (in {pickle_file})"
                        )
                    else:
                        pass
                else:
                    approaching_sd = False
                    appr_distance = (cai.distance_from_sb(
                        self.get_location(), self.approach_leg(cai)) -
                                     cai.get_sd(self.approach_leg(cai)))

                    if appr_distance > 0:
                        approaching_sd = True

                    appr_dict["id"].append(cai.get_id_num())
                    appr_dict["appr_dir"].append(self.approach_leg(cai))
                    appr_dict["timestamp"].append(self.get_utc_timestamp())
                    appr_dict["location"].append(self.get_location())
                    appr_dict["spd"].append(self.get_spd())
                    appr_dict["distance"].append(appr_distance)
                    appr_dict["bearing"].append(self.get_bearing())
                    appr_dict["approaching"].append(approaching_sd)

                    if i == self.gpx_df.last_valid_index() - 1:
                        approach_log_df = pd.DataFrame(appr_dict)
                        approach_log_df.to_csv(csv_file)
                        approach_log_df.to_pickle(pickle_file)
                        print(
                            f"Exported dataframe to CSV ({csv_file}) and Pickle ({pickle_file})"
                        )
        return approach_log_df

    def drive_gpx_stop_bar(self,
                           gpx_filename,
                           use_pickle_file=True) -> pd.DataFrame:
        """ Load the GPX file into a dataframe with timestamp, location, speed, dist to event, bearing, and
        id of approaching intersection

        :param gpx_directory: file directory where .gpx file is for input
        :param gpx_filename: filename of .gpx file (without .gpx)
        :param out_file_directory: where output files are saved to (./out/)
        :param use_pickle_file: default to False, can be faster to load from pickle
        :return: DataFrame with GPX calculated for sro file and saved a CSV and Pickle file of gpx information

        """
        # self.gpx_filepath = gpx_directory + gpx_filename + ".gpx"
        pickle_file = self.out_file_path / (str(gpx_filename) + ".p")
        csv_file = self.out_file_path / (str(gpx_filename) + ".csv")

        approach_sb_log_df = None

        if use_pickle_file and os.path.isfile(pickle_file):
            approach_sb_log_df = pd.read_pickle(pickle_file)
            return approach_sb_log_df
        else:
            # dictionary-> Keys:Values
            appr_dict = {
                "id": [],
                "appr_dir": [],
                "timestamp": [],
                "time_delta": [],
                "location": [],
                "spd": [],
                "distance": [],
                "bearing": [],
                "approaching": []
            }

            for i in tqdm(range(2, self.gpx_df.last_valid_index()),
                          desc="Loading GPX:",
                          unit="GPX Points"):

                self.update_location_simple(i)
                cai = self.get_closest_approaching_intersection()
                if cai is None:
                    if i == self.gpx_df.last_valid_index() - 1:
                        approach_sb_log_df = pd.DataFrame(appr_dict)
                        approach_sb_log_df.to_csv(csv_file)
                        approach_sb_log_df.to_pickle(pickle_file)
                        print(
                            f"exported dataframe to CSV (in {csv_file}) and Pickle (in {pickle_file})"
                        )
                    else:
                        pass
                else:
                    approaching_sd = False
                    appr_distance = (cai.distance_from_sb(
                        self.get_location(), self.approach_leg(cai)) -
                                     cai.get_sd(self.approach_leg(cai)))
                    # print(f'{i}: {appr_distance}ft from {cai.get_name()}')
                    if appr_distance > 0:
                        approaching_sd = True

                    appr_dict["id"].append(cai.get_id_num())
                    appr_dict["appr_dir"].append(self.approach_leg(cai))
                    appr_dict["timestamp"].append(self.get_utc_timestamp())
                    appr_dict["time_delta"].append(self.get_time_step())
                    appr_dict["location"].append(self.get_location())
                    appr_dict["spd"].append(self.get_spd())
                    appr_dict["distance"].append(appr_distance)
                    appr_dict["bearing"].append(self.get_bearing())
                    appr_dict["approaching"].append(approaching_sd)

                    if i == self.gpx_df.last_valid_index() - 1:
                        print("WRITING DICT TO DATAFRAME")
                        approach_sb_log_df = pd.DataFrame(appr_dict)
                        approach_sb_log_df.to_csv(csv_file)
                        approach_sb_log_df.to_pickle(pickle_file)
                        print(
                            f"Exported data frame to CSV ({csv_file}) and Pickle ({pickle_file})"
                        )
                        print(f'ApproachSB_DF:{approach_sb_log_df}')

        return approach_sb_log_df

    def get_street(self, itrsxn: Intersection) -> str:
        """ current Street of intersection approach leg
        """
        apr_leg_index = self.approach_leg(itrsxn)
        return itrsxn.get_name(apr_leg_index)

    def get_info(self, itrsxn: Intersection) -> str:
        """ get ID#, bearing, and name about an intersection

        :param itrsxn: Intersection Object
        :return: string of info in format ID#.Compass_Heading - Intersection Name
        """
        return str(
            f'{itrsxn.get_id_num()}.{self.approach_leg(itrsxn)}-{itrsxn.get_name()}'
        )

    def get_itrsxn_obj_by_id(self, id_num: int) -> Intersection:
        """get intersection object from ID number.

        :param id_num: ID number of intersection object
        :return: intersection object
        """
        mask = self.sro_df['id'] == id_num
        return self.sro_df[mask].iloc[0, 1]

    def get_info_by_id(self, id_num: int, appr_dir) -> str:
        """get info about intersection

        :param id_num: ID # of intersection
        :param appr_dir: approach direction to intersection
        :return: string in format:
            ID#.direction_index - street1 + street2 - event distance string
        """
        mask = self.sro_df['id'] == id_num
        itrsxn = self.sro_df[mask].iloc[0, 1]
        """
        Example: 2.0-YVR+California-35mph-325ft-UTCtime
        Include: direction, Approach Posted Speed
        """
        return str(
            f'{itrsxn.get_id_num()}.{appr_dir}-{itrsxn.get_name()}-{itrsxn.get_sd(appr_dir)}ft'
        )

    @staticmethod
    def find_index(df, i):
        return df.iloc[i]

    @staticmethod
    def t_spd_adjust(d0, spd0, d1, spd1):
        """ Adjusts time of event based on speed of gpx points i and i+1.

        :param d0: distance a t=0
        :param spd0: speed at t=0
        :param d1: distance a t=1
        :param spd1: speed at t=1
        :return: time adjustment from t=0 to event location
        """
        t_adj0 = d0 / spd0
        t_adj1 = d1 / spd1

        d_step = d0 + abs(d1)

        w0 = d0 / d_step
        w1 = abs(d1) / d_step

        sum_t_adj = abs(w0 * t_adj0 + w1 * t_adj1)

        return sum_t_adj / 2.0

    def seek_sd(self, gpx_df, csv_out=True) -> pd.DataFrame:
        """Find event time of Sight Distance(sd) for each intersection traveled through

        :param gpx_df: dataframe of processed GPX file with SRO's events calculated
        :param csv_out: save a CSV file of the intersection and approaches captured
        :return: dataframe of intersection approaches and speed and distance
        """

        # How this method works:
        # i-1 -- Previous timestep gpx point (typically 1 second behind)
        # i   -- current timestep of gpx point
        # i+1 -- next timestep gpx point (typically 1 second ahead)
        # i+2 -- 2nd next timestep gpx point (typically 2 seconds ahead)

        # For each step in the GPS
        #  if speed(i) AND speed(i+1) > 0.4 mph AND distance to sight distance <=
        #   approaching intersection's sight distance + 50ft
        #   AND i-1 AND i are approaching intersection
        #   AND i+1 AND i+2 are not approaching (i.e. just passed event distance)

        # initialize dataframe of sighted intersections and misc. info
        df_dict = {
            "id": [],
            "appr_dir": [],
            "timestamp": [],
            "location": [],
            "spd": [],          # MPH
            "distance": [],     # FEET
            "t_adjust": [],     # SECONDS
            "string_desc": []
        }

        for i in range(1, (gpx_df.last_valid_index() - 3)):
            if gpx_df.spd.iloc[i] > 0.4 and \
                    gpx_df.spd.iloc[i + 1] > 0.4 and \
                    gpx_df.distance.iloc[i + 1] <= \
                    self.get_itrsxn_obj_by_id(gpx_df.id.iloc[i]).get_sd(gpx_df.appr_dir.iloc[i + 1]) + \
                        (gpx_df.time_delta.iloc[i] * (gpx_df.spd.iloc[i] * self.MPHtoFTPS)):
                print("heuristic filter")

                if (gpx_df.approaching.iloc[i - 1]) == True and \
                        (gpx_df.approaching.iloc[i]) == True and \
                        (gpx_df.approaching.iloc[i + 1] == False) and \
                        (gpx_df.approaching.iloc[i + 2] == False):

                    print("approach filter")

                    t_adjust = self.t_spd_adjust(gpx_df.distance.iloc[i], gpx_df.spd.iloc[i],
                                                 gpx_df.distance.iloc[i + 1], gpx_df.spd.iloc[i + 1]
                                                 )


                    # t_adj is less than timestep when dynamic object is close to event distance
                    if t_adjust <= gpx_df.time_delta.iloc[i]:
                        print("t-adjust filter")
                        df_dict["id"].append(gpx_df.id.iloc[i])
                        df_dict["appr_dir"].append(gpx_df.appr_dir.iloc[i])
                        df_dict["timestamp"].append(gpx_df.timestamp.iloc[i])
                        df_dict["location"].append(gpx_df.location.iloc[i])
                        df_dict["spd"].append(gpx_df.spd.iloc[i])
                        df_dict["distance"].append(gpx_df.distance.iloc[i])
                        df_dict["t_adjust"].append(t_adjust)
                        df_dict["string_desc"].append(
                            self.get_info_by_id(gpx_df.id.iloc[i],
                                                gpx_df.appr_dir.iloc[i]))
                        print(
                            f'seek_sd_Info:{self.get_info_by_id(gpx_df.id.iloc[i], gpx_df.appr_dir.iloc[i])}, time adjust:{t_adjust}'
                        )
                    else:
                        pass
            else:
                pass
        if csv_out:
            pd.DataFrame(df_dict).to_csv(
                self.out_file_path / "approaching_intersections.csv"
            )
        return pd.DataFrame(df_dict)

    # TODO: Ensure this works properly for Stop bar data, combine with seek_sd method
    def seek_sb(self, gpx_df, csv_out=True) -> pd.DataFrame:
        """Find event time of Stop Bar (sb)  for each intersection traveled through

        :param gpx_df: dataframe of GPX file
        :param csv_out: save a CSV file of the intersection and approaches captured
        :return: dataframe of intersection approaches and speed and distance
        """

        # See seek_sd for how this method works

        df_dict_sb = {
            "id": [],
            "appr_dir": [],
            "timestamp": [],
            "location": [],
            "spd": [],
            "distance": [],
            "t_adjust": [],
            "string_desc": []
        }

        for i in range(1, (gpx_df.last_valid_index() - 3)):
            if self.get_itrsxn_obj_by_id(gpx_df.id.iloc[i]).distance_from_sb(
                    gpx_df.location.iloc[i + 1], gpx_df.appr_dir.iloc[i + 1]) is None:
                print(f'USING CENTER OF INTERSECTION LOCATION')
                approach_distance = self.get_itrsxn_obj_by_id(
                    gpx_df.id.iloc[i]).get_sd(gpx_df.appr_dir.iloc[i + 1]) + 50
            else:
                print(f'USING STOP BAR LOCATION')
                approach_distance = self.get_itrsxn_obj_by_id(
                    gpx_df.id.iloc[i]).distance_from_sb(gpx_df.location.iloc[i + 1],
                                                        gpx_df.appr_dir.iloc[i + 1])

            if gpx_df.spd.iloc[i] > 0.2 and gpx_df.spd.iloc[i + 1] > 0.2 \
                    and gpx_df.distance.iloc[i + 1] <= approach_distance <= gpx_df.distance.iloc[i]:

                if True:
                    """
                    calculate exact time based on speed. using i and i+1. 
                    
                    For point i and i+1, the exact time adjustment is weighted
                    based on the speeds at these two point's in time. This shifts the exact position
                    of the car to the most accurate time it was at the calculated sight distance
                    
                    store that time in UTC in DF
                    """

                    t_adjust = self.t_spd_adjust(gpx_df.distance.iloc[i], gpx_df.spd.iloc[i],
                                                 gpx_df.distance.iloc[i + 1], gpx_df.spd.iloc[i + 1]
                                                 )

                    if t_adjust <= 1:
                        df_dict_sb["id"].append(gpx_df.id.iloc[i])
                        df_dict_sb["appr_dir"].append(gpx_df.appr_dir.iloc[i])
                        df_dict_sb["timestamp"].append(gpx_df.timestamp.iloc[i])
                        df_dict_sb["location"].append(gpx_df.location.iloc[i])
                        df_dict_sb["spd"].append(gpx_df.spd.iloc[i])
                        df_dict_sb["distance"].append(gpx_df.distance.iloc[i])
                        df_dict_sb["t_adjust"].append(t_adjust)
                        df_dict_sb["string_desc"].append(
                            self.get_info_by_id(gpx_df.id.iloc[i],
                                                gpx_df.appr_dir.iloc[i]))

                        # print(self.utc_to_timestamp(df.timestamp.iloc[i]))
                        print(
                            f'info:{self.get_info_by_id(gpx_df.id.iloc[i], gpx_df.appr_dir.iloc[i])}'
                        )

                    else:
                        pass
            else:
                pass
        if csv_out:
            pd.DataFrame(df_dict_sb).to_csv(
                self.out_file_path / "approaching_intersections_Stopbar.csv")
        return pd.DataFrame(df_dict_sb)

    # TODO: create new method
    # def get_info_at_timestamp(timestamp)
    #   return intersection ID, approach, spd, distance, compass direction, lat, lon


class Vehicle(DynamicRoadObject):
    pass



# !/usr/bin/env python
# coding: utf-8

import csv
import os
from datetime import datetime, timezone
from pathlib import Path, PurePath

import geopy
import gpxpy
import gpxpy.gpx

import pandas as pd
from geopy.distance import geodesic, Distance
import lxml
from lxml import etree
from tqdm import tqdm

from ssoss.static_road_object import Intersection
from ssoss.motion_road_object import GPXPoint


class ProcessRoadObjects:

    def __init__(self,
                 signals_filename: str,
                 gpx_filename: str,
                 in_dir_path=PurePath("./in"),
                 in_gpx_dir_path=PurePath("/gpx_video/"),
                 out_dir_path=PurePath("./out/")
                 ):
        """ Class to process Road Object files. Using January 1st 1970 as time epoc

        :in_dir_path - defaults to ./in/
        :in_gpx_dir_path = defaults to gpx_video/ folder for both .gpx and video files
        :out_dir_path = defaults to ./out/
        """

        self.intersection_listDF = None
        self.intersection_load = None
        self.date_format = "%m-%d-%Y--%H-%M-%S.%f-%Z"
        self.pretty_datetime_format = "%y-%m-%d %H:%M:%S"

        self.in_dir_path = in_dir_path
        self.in_gpx_dir_path = in_gpx_dir_path
        self.out_dir_path = out_dir_path

        # init variables
        self.intersection_filename = signals_filename
        self.gpx_filename = gpx_filename
        self.pickle_file = ''
        self.gpx_file = ''
        self.csv_file = None
        self.gpxDF = ''
        self.gpx_listDF = None
        self.gDF_pickle_file = None

        self.sum_time_gap = 0.0
        self.sum_total_points = 0.0

        # scafold directory structure if not present
        gpx_video_dir = self.in_dir_path / self.in_gpx_dir_path
        p = Path(str(gpx_video_dir))

        if not p.is_dir():
            os.makedirs(gpx_video_dir, exist_ok=True)
            os.makedirs(self.out_dir_path, exist_ok=True)
            print(f"Created ./{self.in_dir_path} and ./{gpx_video_dir}/ directories for input files")
            print(f"Load formated CSV signal file to {self.in_dir_path}"
                  f"Load GPX dynamic route file to ./{gpx_video_dir}"
                  f"Load Video recording files to ./{gpx_video_dir}"
                  )

        all_intersections_df = self.load_intersection_csv(signals_filename)
        gpx_df = self.load_gpx_to_obj_df(gpx_filename, use_pickle=False)



    @staticmethod
    def speed_calc(point1, point2, t1, t2) -> float:
        """
        Calculates Meters Per Second speed between two point and time objects
        """
        # dist = geo.distance(point1.latitude, point1.longitude, point2.latitude, point2.longitude)
        dist = geodesic(point1, point2).meters
        time = (t2 - t1).total_seconds()  # timedelta converted to float

        if time > 0:
            speed = dist / time
            if speed < 50:  # 50 meters/sec threshold for accurate speed
                return speed
            else:
                return 0.0
        else:
            return 0.0

    def get_intersection_object_by_id(self, intersection_id):
        return self.intersection_listDF.iloc[intersection_id-1, 1]

    def intersection_frame_description(self, sro_id, b_index, distance, ts, desc_type="filename"):
        i_obj = self.get_intersection_object_by_id(sro_id)

        i_id = sro_id
        i_bearing = b_index
        i_compass_bearing = i_obj.get_bearing_str(b_index)
        i_name = i_obj.get_name()
        i_name_one = i_obj.get_name(0)
        i_name_two = i_obj.get_name(1)
        i_sd = i_obj.get_sd(b_index)
        i_dist = distance
        ts_utc = round(ts, 3)
        dt_temp = datetime.fromtimestamp(ts, tz=None)
        date_time = dt_temp.strftime("%a, %b %-d %Y at %I:%-M %p")

        #  don't use "/" when building filename string
        if desc_type == "filename":
            filename_desc = f'{i_id}.{i_bearing}-{i_name}-{i_sd}-{ts_utc}'
            return filename_desc
        elif desc_type == "label":
            label = f'{i_compass_bearing} approach of {i_name_one} and {i_name_two} (#{sro_id}) at ~{i_sd} ft on {date_time}'
            return label


    def load_intersection_csv(self, intersection_filename: str) -> pd.DataFrame:
        """ Loads CSV file into Intersection Class DataFrame

        :param intersection_filename: name of CSV file for loading (leave off .csv)
        :return: dataframe of intersections objects in each row
        """

        csv_intersection_file = PurePath(self.in_dir_path, intersection_filename + ".csv")
        self.intersection_load = {"id": [], "intersection_obj": []}

        with open(csv_intersection_file, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            line_count = 0
            for row in csv_reader:
                self.intersection_load["id"].append(int(row[0]))
                self.intersection_load["intersection_obj"].append(
                    Intersection(
                        # id_num
                        int(row[0]),
                        # name1(N/S), name2(E/W)
                        tuple((str(row[1]),
                               str(row[2]))),
                        # center of intersection location
                        geopy.Point(float(row[3]),
                                    float(row[4])),
                        # spd_N, spd_E, spd_S, spd_W
                        tuple((int(row[5]), int(row[6]), int(row[7]),
                               int(row[8]))),
                        # bearing_N, bearing_E, bearing_S, bearing_W
                        tuple((float(row[9]), float(row[10]), float(row[11]),
                               float(row[12]))),
                        # NB Stop bar. center line Point(lat, lon), shoulder Point(lat, lon)
                        tuple((geopy.Point(row[13], row[14]),
                               geopy.Point(row[15], row[16]))),
                        # EB Stop bar. center line Point(lat, lon), shoulder Point(lat, lon)
                        tuple((geopy.Point(row[17], row[18]),
                               geopy.Point(row[19], row[20]))),
                        # SB Stop bar. center line Point(lat, lon), shoulder Point(lat, lon)
                        tuple((geopy.Point(row[21], row[22]),
                               geopy.Point(row[23], row[24]))),
                        # WB Stop bar. center line Point(lat, lon), shoulder Point(lat, lon)
                        tuple((geopy.Point(row[25], row[26]),
                               geopy.Point(row[27], row[28])))
                    ))

                line_count += 1

            self.intersection_listDF = pd.DataFrame(self.intersection_load)
            print(
                f"Processed {line_count} lines of CSV file for a total of {len(self.intersection_listDF.index)} intersections."
            )
        return self.intersection_listDF

    def set_gpx_ver(self):
        gpx_string = etree.tostring(etree.parse(self.gpx_file), encoding=str)
        gpx_ver_10 = gpx_string.find("http://www.topografix.com/GPX/1/0")
        gpx_ver_11 = gpx_string.find("http://www.topografix.com/GPX/1/1")
        if gpx_ver_10 < 0 and gpx_ver_11 > 0:
            self.gpx_ver = "1.1"
            return self.gpx_ver
        else:
            self.gpx_ver = "1.0"
            return self.gpx_ver

    def load_gpx_to_obj_df(self, gpx_filename: str, gpx_ver = "1.0", use_pickle=True) -> pd.DataFrame:
        """ Loads GPX file into point objects and returns a dataframe of all the points
        """

        self.gpx_filename = gpx_filename
        self.gpx_ver = gpx_ver
        self.gpx_file = self.in_dir_path / self.in_gpx_dir_path / (gpx_filename + ".gpx")
        self.pickle_file = self.out_dir_path / (gpx_filename + ".pkl")
        self.csv_file = self.out_dir_path / (gpx_filename + ".csv")

        gpx_load = {"gpx_pt": []}

        # initialize starting variables if GPX v1.1 needs speed calcs
        pnt1 = geopy.Point()
        t1 = datetime.now(timezone.utc)

        if use_pickle and Path(self.pickle_file).is_file():
            self.gpxDF = pd.read_pickle(self.pickle_file)
            print(
                f"Loaded Pickle file {self.pickle_file} into Dataframe with {self.gpxDF.last_valid_index()} rows"
            )
        else:
            print(
                f"Using GPX file: {self.gpx_file}"
            )

        self.gpx_ver = self.set_gpx_ver()

        gpx_file_ref = open(self.gpx_file, "r")
        # use GPX v1.0, assume speed data otherwise will calculate
        # if speed is captured in GPX v1.1 as extension data, change version to 1.1
        # TODO: set timezone with lat, lon coordinates:
        #  (https://stackoverflow.com/questions/15742045/getting-time-zone-from-lat-long-coordinates)


        gpx = gpxpy.parse(gpx_file_ref, version=self.gpx_ver)
        pt_count = 0
        for track in gpx.tracks:
            for segment in track.segments:
                for point in segment.points:
                    if self.gpx_ver == "1.1":
                        extra_data = {}
                        extension_data = {
                            lxml.etree.QName(child).localname: float(child.text)
                            for child in point.extensions[0]
                        }
                        for k, v in extension_data.items():
                            extra_data[k] = v

                        point.speed = extra_data["speed"]

                    p = geopy.Point(
                        latitude=point.latitude,
                        longitude=point.longitude,
                    )

                    if point.speed is not None:
                        pass    # GPX v1.0 includes speed in track, v1.1 can include in extension data
                    elif point.speed is None:  # GPX v1.1 speed calculation
                        if pt_count == 0:
                            point.speed = 0
                        else:
                            pnt0 = pnt1
                            pnt1 = p
                            t0 = t1
                            t1 = point.time

                            point.speed = self.speed_calc(pnt0, pnt1, t0, t1)

                    gpx_load["gpx_pt"].append(
                        GPXPoint(
                            pt_count,
                            str(point.time),  # convert ISO format point.time to a string for timestamp conversion
                            tuple(p),  # tuple of point, lat and lon
                            point.speed
                                 )
                    ) #  tuple for point for csv file output?.
                    pt_count += 1


        self.gpx_listDF = pd.DataFrame(gpx_load)
        self.gpx_listDF.to_pickle(self.pickle_file)
        self.gpx_listDF.to_csv(self.csv_file)
        print(
            f"Processed {pt_count} points of GPX file."
        )
        self.update_gpx_points()

        self.gpx_summary()


        return self.gpx_listDF

    def update_gpx_points(self):
        """
        updates gpx_prev_point and gpx_next_point as objects after the initial points
        are loaded
        """
        # simplify variable names
        last_index = self.gpx_listDF.last_valid_index()
        gpx_df = self.gpx_listDF
        cuml_d = 0.0

        # edge cases
        gpx_df.iloc[0, 0].set_next_gpx_point(gpx_df.iloc[1, 0])


        for i in range(1, last_index):  # populate prev_gpx_points
            gpx_df.iloc[i, 0].set_prev_gpx_point(gpx_df.iloc[i-1, 0])
        for i in range(last_index-1, 0, -1):  # reversed range to populate next_gpx_points
            gpx_df.iloc[i, 0].set_next_gpx_point(gpx_df.iloc[i+1, 0])

        # edge cases
        gpx_df.iloc[0, 0].set_next_gpx_point(gpx_df.iloc[1, 0])
        gpx_df.iloc[last_index, 0].set_prev_gpx_point(gpx_df.iloc[last_index - 1, 0])

        # repopulate prev_gpx_point objects so next_gpx_points are not NONE
        for i in range(1, last_index+1):
            current_gpx_pt = gpx_df.iloc[i, 0]
            current_gpx_pt.set_prev_gpx_point(gpx_df.iloc[i-1, 0])
            self.sum_time_gap += current_gpx_pt.get_prev_timedelta()
            cuml_d += current_gpx_pt.distance_to(current_gpx_pt.get_prev_gpx_point().get_location())
            current_gpx_pt.set_cumulative_distance(cuml_d)


        # edge cases again
        gpx_df.iloc[0, 0].set_next_gpx_point(gpx_df.iloc[1, 0])


        # display progress bar for calculating time-consuming backflow function
        for i in tqdm(range(len(gpx_df.index))):
            gpx_df.iloc[i, 0].backflow(self.intersection_listDF)

    def get_start_timestamp(self):
        return self.gpx_listDF.iloc[0, 0].get_timestamp()

    def get_end_timestamp(self):
        last_index = self.gpx_listDF.last_valid_index()
        return self.gpx_listDF.iloc[last_index, 0].get_timestamp()

    def intersection_sight_distance_timestamps(self):
        gpx_df = self.gpx_listDF
        all_intersections = self.intersection_listDF

        intersection_sd = []  # store intersection id & index in list
        intersection_ts = []  # store timestamps in list

        for point in range(len(gpx_df.index)):
            intersections_info = gpx_df.iloc[point, 0].get_intersection_approach_list()
            if intersections_info:
                p = gpx_df.iloc[point, 0]
                intersection_id, bearing_index, dist, approach = zip(*intersections_info)
                for item in range(len(list(intersection_id))):
                    sro_id = int(list(intersection_id)[item])
                    b_index = int(list(bearing_index)[item])
                    d_current = list(dist)[item]

                    approach_intersection = all_intersections.iloc[sro_id-1, 1]  # shift down by 1 for list
                    approach_intersection_sight_distance = approach_intersection.get_sd(b_index)

                    prev_p = p.get_prev_gpx_point()
                    next_p = p.get_next_gpx_point()
                    if prev_p is not None:
                        d_prev = p.get_prev_gpx_point().distance_to(approach_intersection.get_location())
                    if next_p is not None:
                        d_next = p.get_prev_gpx_point().distance_to(approach_intersection.get_location())

                    if p.h_prev_and_current_before_next(approach_intersection, b_index) and \
                       p.h_next_less_than_current(approach_intersection, b_index):

                        t_acc = p.t_to_approach_acc(approach_intersection, b_index)
                        print(
                            f"Signal #{sro_id}.{b_index} at {approach_intersection_sight_distance} ft")

                        t_shift_acc = p.get_timestamp() + t_acc
                        intersection_sd.append(self.intersection_frame_description(sro_id, b_index, d_current, t_shift_acc))
                        intersection_ts.append(t_shift_acc)

        z = zip(intersection_sd, intersection_ts)
        id_ts = list(z)
        ret = sorted(id_ts, key=lambda x: x[1])

        return ret


    @staticmethod
    def hr_min_sec(sec):
        if sec < 60:
            return f'{sec} seconds'
        elif sec < 3600:
            min = int(sec/60)
            sec_remain = round(sec - min * 60, 2)
            return f'{min}:{sec_remain} (MM:SS.ss)'
        elif sec >= 3600:
            hr = int(sec/3600)
            min = int(sec/60)
            sec_remain = round(sec - min * 60, 2)
            return f'{hr}:{min}:{sec_remain} (HH:MM:SS.ss)'

    @staticmethod
    def simplify_distance(d_ft):
        if d_ft < 5280:
            return f'{round(Distance(feet=d_ft).ft, 2)} feet'
        else:
            return f'{round(Distance(feet=d_ft).miles, 2)} miles'

    def gpx_summary(self):
        gpx_df = self.gpx_listDF
        last_index = gpx_df.last_valid_index()
        self.sum_total_points = last_index + 1  # pd dataframe starts at zero
        avg_time_gap = round(self.sum_time_gap / self.sum_total_points, 2)
        tot_sec = round(self.get_end_timestamp() - self.get_start_timestamp(), 2)
        tot_distance = gpx_df.iloc[last_index, 0].get_cumulative_distance()

        # display values
        width = int(70)
        title = "GPX SUMMARY"
        symbol = "-"

        summary = f"""
        {symbol * width}
        {" " * (int(width/2)-int(len(title)/2))}{title}
        {symbol * width}
        # GPX File:: {self.in_dir_path}/{self.gpx_file}
        # Using GPX version: {self.gpx_ver}
        # Start time: {datetime.fromtimestamp(self.get_start_timestamp(), tz=None)}
        # End time:   {datetime.fromtimestamp(self.get_end_timestamp(), tz=None)}
        # Total duration: {self.hr_min_sec(tot_sec)}
        # Total distance: {self.simplify_distance(tot_distance)}
        # Number of data points: {self.sum_total_points}
        # Avg. Time Gap between data points: {avg_time_gap} Seconds
        {symbol * width}
        """
        # TODO:
        # difference in GPX and Video file start times and lengths of times
        # Number of images captured:
        # Number of intersections captures: X/ Total intersections (xx.x%)
        # Number of approaches captured
        # Approaches captures for duration of GPX file and Video File -> (images/time) (productivity ratio)
        # -------------------------------------------------------------------------------

        print(summary)










# !/usr/bin/env python
# coding: utf-8

import math

import geopy
import geopy.distance


class StaticRoadObject:

    def __init__(self, id_num: int, name: str, obj_type: type,
                 ctr_pt: geopy.Point, spd_sd: dict):
        """initializes values related to static road objects

        :param id_num: Identification number (int)
        :param name: name of street object is located on
        :param obj_type: type of static object, ie. sign, intersection, etc.
        :param ctr_pt: Geopy Point object of lat, lon, altitude
        :param spd_sd: speed [key] and sight distance[value] of object to be viewed (distance in ft)
        """

        self.id_num = id_num
        self.name = name
        self.obj_type = obj_type
        self.ctr_pt = ctr_pt
        self.pt = geopy.Point(ctr_pt.latitude, ctr_pt.longitude)  # removes elevation for dist calcs
        self.spd_sd = spd_sd

    def get_id_num(self) -> int:
        return int(self.id_num)

    def get_name(self) -> str:
        return self.name

    def get_location(self) -> geopy.Point:
        return self.ctr_pt

    def get_obj_type(self):
        return self.obj_type

    def get_sd(self):
        return self.spd_sd.values()[0]  # first/only distance (feet) to have clear view of static object.


# class for speed limit signs
class SpeedLimitSign(StaticRoadObject):
    pass


# class for radar signs that may need greater sight distance
class RadarSpeedFeedbackSign(StaticRoadObject):
    pass


# Generic class for stop signs, street name signs, curve warning signs, etc.
class TrafficControlSign(StaticRoadObject):
    pass


class Intersection(StaticRoadObject):
    """Extends Static Road Objects with additional variables and methods related to intersection objects
    """

    def __init__(self,
                 id_num,
                 name: tuple,
                 ctr_pnt: geopy.Point,
                 spd: tuple,
                 bearing: tuple,
                 stop_bar_nb=(0, 0),
                 stop_bar_eb=(0, 0),
                 stop_bar_sb=(0, 0),
                 stop_bar_wb=(0, 0)):
        """initialize variables of intersection class, mostly stored as tuples in (North, East, South, West) format

        :param id_num:
        :param name: tuple of the name of two streets intersecting ((N/S, E/W))
        :param ctr_pnt: center lat, lon of intersection as geopy.Point object
        :param spd: tuple of posted speed limit values for each approach ((N,E,S,W))
        :param bearing: tuple of compass bearing for each approach ((N,E,S,W))
        :param stop_bar_nb: tuple of two geopy Points (both with (lat,lon)) for NB approach stop bar*
        :param stop_bar_eb: tuple of two geopy Points (both with (lat,lon)) for EB approach stop bar*
        :param stop_bar_sb: tuple of two geopy Points (both with (lat,lon)) for SB approach stop bar*
        :param stop_bar_wb: tuple of two geopy Points (both with (lat,lon)) for WB approach stop bar*

        *Note: left point (inside lane) is [0] in tuple, right point (right turn lane) is [1] in tuple

        """

        self.stop_bar = None
        self.spd = spd
        self.ctr_pnt = ctr_pnt
        self.bearing = bearing
        self.spd_sd = {
            -999: 0,
            20: 175,
            25: 215,
            30: 270,
            35: 325,
            40: 390,
            45: 460,
            50: 540,
            55: 625,
            60: 715
        }
        #  default to lowest distance if speed not found in spd_sd dict.
        self.sd = tuple((self.spd_sd.get(self.spd[0], 175),
                         self.spd_sd.get(self.spd[1], 175),
                         self.spd_sd.get(self.spd[2], 175),
                         self.spd_sd.get(self.spd[3], 175),
                         ))

        self.stop_bar_d = tuple((stop_bar_nb, stop_bar_eb,
                                 stop_bar_sb, stop_bar_wb
                                 ))

        StaticRoadObject.__init__(self, id_num, name, Intersection, ctr_pnt,
                                  self.spd_sd)

    # TODO: Convert this to dictionary from input file, not hard coded values
    # @staticmethod
    # def spd_mph2sd(spd_mph) -> int:
    #     """ based on CA-MUTCD intersection sight distances based on speed
    #     """
    #     if spd_mph == 20:
    #         return 175
    #     elif spd_mph == 25:
    #         return 215
    #     elif spd_mph == 30:
    #         return 270
    #     elif spd_mph == 35:
    #         return 325
    #     elif spd_mph == 40:
    #         return 390
    #     elif spd_mph == 45:
    #         return 460
    #     elif spd_mph == 50:
    #         return 540
    #     elif spd_mph == 55:
    #         return 625
    #     elif spd_mph == 60:
    #         return 715
    #     else:
    #         return 0

    def get_name(self, index=-1) -> str:
        """ return name of intersection from ID #.
        """
        if index == -1:
            return str(self.name[0]) + "+" + str(self.name[1])
        elif index == 0 or index == 2:
            return str(self.name[0])
        elif index == 1 or index == 3:
            return self.name[1]

    def get_bearingT(self) -> tuple:
        return self.bearing

    def get_bearing(self, direction: int) -> float:
        return self.bearing[direction]

    @staticmethod
    def get_bearing_str(direction: int) -> str:
        if direction == 0:
            return "NB"
        elif direction == 1:
            return "EB"
        elif direction == 2:
            return "SB"
        elif direction == 3:
            return "WB"

    def get_sdT(self) -> tuple:
        return self.sd

    def get_sd(self, direction) -> float:
        """ get sight distance of this static object based on compass direction/heading
        """
        if direction == "min":
            return float(
                self.spd_sd[20] - self.spd_sd[20] * .5
                # allows for i+2 on GPX comparison to calculate when going 30mph (for 25mph zone)
            )
        elif direction == "max":
            return float(
                self.spd_sd[60] * 1.5  # allows for i+2 on GPX approaching algo for 60mph.
            )
        else:
            return round(float(self.sd[direction]))

    def get_sb_distance(self, direction) -> float:
        """ NOT SURE WHAT THIS DOES
        get stop bar distance of intersection object based on compass direction/heading
        :param direction:
        :return: float od distance in feet from stop bar to event distance
        """
        return float(self.stop_bar_d[direction])

    def get_info(self) -> str:
        """ get id # and name of intersection object
        """
        return str(f'{self.get_id_num()}-{self.get_name()}')

    def distance_to_sb(self, dynamic_pt: geopy.Point,
                       direction: int) -> float:
        """ Calculates distance from point to stop bar if stop bar location is provided
        NOTE: If Stop bar distance not provided, defaults to  center of the intersection point

        :param dynamic_pt: geopy point object of dynamic object
        :param direction: index for heading of dynamic object
        :return: distance in feet as float from dynamic point to stop bar of approaching intersection
        """

        # length of stop bar
        a = geopy.distance.distance(self.stop_bar_d[direction][0],
                                    self.stop_bar_d[direction][1]).ft

        # if length of stop bar is zero, use center of intersection point.
        if a <= 0 or self.stop_bar is None:
            dis_to_ctr = geopy.distance.distance(self.pt, dynamic_pt).ft
            return dis_to_ctr
        else:
            # distance from dynamic point to inside lane stop bar point
            b = geopy.distance.distance(self.stop_bar[direction][0], dynamic_pt).ft
            # distance from dynamic point to outside lane stop bar point
            c = geopy.distance.distance(self.stop_bar[direction][1], dynamic_pt).ft

        # calculate perpendicular distance between stop bar and dynamic object
        if a is not None and b is not None and c is not None:
            s = (a + b + c) / 2
            dis_to_sb = 2. * math.sqrt(abs(s * (s - a) * (s - b) *
                                           (s - c))) / a
            return dis_to_sb
        else:
            return 0


def main():
    print("Initialize and Load Static Road Object File")
    print("Copy CSV file into ./in/ folder, where ./ is the current directory")


if __name__ == "__main__":
    main()



# !/usr/bin/env python
# coding: utf-8
import math
from datetime import datetime, timezone
from datetime import timedelta
from operator import attrgetter, itemgetter
from pathlib import PurePath

import dateutil
import geopy
import gpxpy.geo as gpxgeo
import numpy as np
import pandas as pd

from ssoss.static_road_object import StaticRoadObject, Intersection


class GPXPoint:
    """class for GPX points to calculate necessary distances and positions to other objects
    """

    def __init__(self, id_num: int, t, p: geopy.Point, spd: float):
        """ TODO: do we need a timezone parameter?

        :param id_num: count number of point in GPX file
        :param t: timestamp of gpx point
        :param pt: geopy point, longitude and latitude of point
        :param spd: speed in ft/sec at that point
        """
        # constants for unit conversions
        self.MStoMPH = 2.23694
        self.FTPStoMPH = 0.681818
        self.MPHtoFTPS = 1 / self.FTPStoMPH
        self.MStoFTPS = self.MStoMPH * self.MPHtoFTPS


        # initial variables from GPX file

        self.id = id_num
        t_temp = (dateutil.parser.isoparse(t))
        self.t = t_temp.replace(tzinfo=timezone.utc).timestamp()
        self.p = geopy.Point(p[0], p[1])  # elevation not supported
        self.spd = spd

        # calculated variables from backflow function
        self.prev_gpx_point = None  # GPX Class Object
        self.next_gpx_point = None  # GPX Class Object
        self.bearing = None

        # advanced calculated
        self.intersection_approach_list = None
        self.cumulative_distance = 0.0
        #self.closest_intersection_list = None                # sorted list
        #self.approaching_intersection_list = None            # dict:  #   True/False
        #self.closest_approaching_intersection = None    # sorted list
        # self.approach_leg(intersection_ID)        # index value
        # self.sight_distance(leg, intersection_id) # distance in feet



    def get_id(self) -> int:
        return self.id

    def get_timestamp(self) -> datetime:
        return self.t

    def get_prev_timedelta(self) -> float:
        return self.t - self.prev_gpx_point.get_timestamp()

    def get_next_timedelta(self) -> float:
        return self.next_gpx_point.get_timestamp() - self.t

    def get_location(self) -> geopy.Point:
        return self.p

    def get_intersection_approach_list(self):
        return self.intersection_approach_list

    def distance_to(self, p1) -> geopy.distance:
        return geopy.distance.distance(p1, self.p).ft

    def get_dist_between_points(self, p1, p2) -> geopy.distance:
        return geopy.distance.distance(p1, p2).ft

    def get_cumulative_distance(self):
        return self.cumulative_distance  # Feet

    def get_speed(self, units="ft_per_sec") -> float:
        """
        speed from gpx is meters/second
        :return: speed in ft/second
        """
        if units == "ft_per_sec":
            return self.spd * self.MStoFTPS

    def get_prev_gpx_point(self):
        return self.prev_gpx_point  # GPX Point Object

    def get_next_gpx_point(self):
        return self.next_gpx_point  # GPX Point Object

    def set_prev_gpx_point(self, o):
        self.prev_gpx_point = o

    def set_next_gpx_point(self, o):
        self.next_gpx_point = o

    def set_cumulative_distance(self, d):
        self.cumulative_distance = d

    def is_intersection_in_next_point(self, i, b) -> bool:
        out_bool = False
        next_intersection_appr_list = self.next_gpx_point.get_intersection_approach_list()

        if next_intersection_appr_list:
            id_list, b_index, dist, appr = zip(*next_intersection_appr_list)
            for x in range(len(list(id_list))):
                if i == id_list[x] and b == b_index[x]:
                    out_bool = True

        return out_bool

    def approaching(self, sro: StaticRoadObject) -> bool:
        if self.prev_gpx_point is None:
            return False
        elif self.distance_to(sro.get_location()) < self.prev_gpx_point.distance_to(sro.get_location()):
            return True
        else:
            return False

    def get_bearing(self) -> float:
        if self.prev_gpx_point is None:
            return 0
        else:
            prev_lat = self.prev_gpx_point.get_location().latitude
            prev_lon = self.prev_gpx_point.get_location().longitude
            cur_lat = self.get_location().latitude
            cur_lon = self.get_location().longitude

            self.bearing = gpxgeo.get_course(prev_lat, prev_lon, cur_lat, cur_lon)
            return self.bearing

    def calc_bearing_diff(self, m: float) -> float:
        n = self.get_bearing()
        b_diff = min(abs(n - m), abs(360 - n + m), abs(360 - m + n))
        return b_diff

    def get_approach_leg(self, intersection: Intersection, index_out=True):
        """determines the approach leg of the intersection based on compass headings

        :param intersection: intersection object to find approach leg of
        :param index_out: when True returns integer index, false returns string.
        :return: index of leg being approached of intersection based on bearing
               "True" index_out: 0 - North, 1 - East, 2 - South, 3 - West
               "False" index_out returns strings:  "NB","EB", "SB", "WB".
        """

        pt_to_intersection_diff = [
            self.calc_bearing_diff(intersection.get_bearing(0)),
            self.calc_bearing_diff(intersection.get_bearing(1)),
            self.calc_bearing_diff(intersection.get_bearing(2)),
            self.calc_bearing_diff(intersection.get_bearing(3))
        ]

        approach_leg_index = np.argmin(pt_to_intersection_diff)
        approach_leg_string = ""

        if intersection is None:
            return None
        elif index_out:
            return int(approach_leg_index)
        else:
            if approach_leg_index == 0:
                approach_leg_string = "NB"
            elif approach_leg_index == 1:
                approach_leg_string = "EB"
            elif approach_leg_index == 2:
                approach_leg_string = "SB"
            elif approach_leg_index == 3:
                approach_leg_string = "WB"
            return approach_leg_string

    def t_to_approach_simple(self, approaching_intersection: Intersection, b_index: int) -> float:
        d = self.distance_to(approaching_intersection.get_location()) - approaching_intersection.get_sd(b_index)
        if self.get_speed() > 0:
            return d / self.get_speed()
        else:
            return 0

    def acceleration(self) -> float:

        if self.get_next_gpx_point() is not None and self.get_next_timedelta is not None:
            t_delta = self.get_next_timedelta()
            v_initial = self.get_speed()
            v_final = self.get_next_gpx_point().get_speed()
            return (v_final - v_initial) / t_delta

        else:
            return 0.0

    def t_to_approach_acc(self, approaching_intersection: Intersection, b_index: int) -> float:
        approach_i_sd = approaching_intersection.get_sd(b_index)
        d_sd = self.distance_to(approaching_intersection.get_location()) - approach_i_sd

        # radical: sqrt(v^2 - 4 * acc * d_sd)
        if self.get_speed()**2 > 4 * self.acceleration() * d_sd:
            radical = math.sqrt(self.get_speed()**2 - 4 * self.acceleration() * d_sd)
        else:
            radical = 0
        # denominator: 2 * acc
        denominator = 2*self.acceleration()

        if denominator == 0 or d_sd <= 0:
            return self.t_to_approach_simple(approaching_intersection, b_index)
        else:
            t_acc_pos = (-self.get_speed() + radical) / denominator
            t_acc_neg = (-self.get_speed() - radical) / denominator
            return min(abs(t_acc_neg), abs(t_acc_pos))

    def backflow(self, sro_df: pd.DataFrame):
        """
        after initial GPX points loaded, used intersection dataframe objects to calculate
        values of interest.

        :param sro_df: static road object loaded as dataframe.
        :return:
        """
        # empty lists
        intersection_id = []
        approach_leg = []
        dist = []
        approaching = []

        for index, row in sro_df.iterrows():
            intersection = row["intersection_obj"]
            distance_to_intersection = self.distance_to(intersection.get_location())
            # TODO: consider add min trim also?
            if distance_to_intersection > intersection.get_sd("max"):  # only load relevant distances
                pass
            else:
                intersection_id.append(intersection.get_id_num())
                approach_leg.append(self.get_approach_leg(intersection))
                dist.append(distance_to_intersection)
                approaching.append(self.approaching(intersection))

        temp_all_lists = zip(intersection_id, approach_leg, dist, approaching)
        temp_sort_distance = sorted(temp_all_lists, key=itemgetter(2))  # sort by item 2/distance
        temp_sort_approaching = sorted(temp_sort_distance, key=itemgetter(3), reverse=True)  # sort by item 3/approaching boolean
        only_approaching_intersections = filter(lambda x: x[3] is True, temp_sort_approaching)  # filter out intersections not approached
        self.intersection_approach_list = list(only_approaching_intersections)

        return;
    #  heuristic: d0 and d1 upstream from sight distance (X), d2 downstream of sight distance(X)
    #  -d0-------d1----X-----d2------0
    def h_prev_and_current_before_next(self, approaching_intersection: Intersection, b_index: int) -> bool:
        h_flag = False
        p_prev = self.get_prev_gpx_point()
        p_next = self.get_next_gpx_point()

        approach_i_sd = approaching_intersection.get_sd(b_index)

        d0 = p_prev.distance_to(approaching_intersection.get_location())
        d1 = self.distance_to(approaching_intersection.get_location())
        if p_next is not None:
            d2 = p_next.distance_to(approaching_intersection.get_location())
        else:
            d2 = 0

        if d0 >= approach_i_sd:
            if d1 >= approach_i_sd:
                if d2 <= approach_i_sd:
                    h_flag = True

        return h_flag

    #  heuristic: d2 less than d1
    #  -d1---X----d2------0
    def h_next_less_than_current(self, approaching_intersection: Intersection, b_index: int) -> bool:
        h_flag = False
        p_next = self.get_next_gpx_point()
        approach_i_sd = approaching_intersection.get_sd(b_index)
        d1 = abs(self.distance_to(approaching_intersection.get_location()))
        if p_next is not None:
            d2 = abs(p_next.distance_to(approaching_intersection.get_location()) - approach_i_sd)
        else:
            d2 = 0

        if d2 <= d1:
            h_flag = True

        return h_flag



    # TODO: consider developing heuristic based on --d0----d1----X---0  without a next_d available
    # TODO: consider developing heuristic based on narrowing based on approach angle








from ssoss.dynamic_road_object import *
from ssoss.motion_road_object import *
from ssoss.process_road_objects import *
from ssoss.process_video import *
from ssoss.static_road_object import *
import importlib.metadata

__version__ = importlib.metadata.version("ssoss")




# !/usr/bin/env python
# coding: utf-8
import os
from pathlib import PurePath, Path

import dateutil
import numpy as np

from tqdm import tqdm
from datetime import timedelta, timezone, datetime

import cv2


class ProcessVideo:

    def __init__(self, video_filename, in_dir_path=PurePath("./in/"), in_video_dir_path=PurePath("gpx_video/")):
        """Process video files using methods to extract range of frames,
        extract frame at precise UTC time, or generate gif from selection of images.
        Note: For syncing video and GPX, use set_start_utc()

        :param in_dir_path: filename of video to be processed (include video extension (.mov, .mp4, etc)
        """

        self.DATE_FORMAT = '%m-%d-%Y--%H-%M-%S.%f-%Z'  #ISO 8601 format
        self.video_filepath = in_dir_path / in_video_dir_path / video_filename  # video path and filename
        self.video_filename = video_filename
        self.image_out_path = PurePath("./out/frames/")
        self.fps = self.get_fps()
        self.frame_count = self.get_frame_count()
        self.duration = self.get_duration()
        self.start_time = 0
        self.capture = ""


    def set_start_utc(self, video_start_time):
        self.start_time = video_start_time
        return None

    def get_start_timestamp(self):
        return self.start_time

    def get_fps(self):
        self.capture = cv2.VideoCapture(str(self.video_filepath))
        self.fps = self.capture.get(cv2.CAP_PROP_FPS)
        self.capture.release()
        return self.fps

    def get_frame_count(self):
        self.capture = cv2.VideoCapture(str(self.video_filepath))
        self.frame_count = self.capture.get(cv2.CAP_PROP_FRAME_COUNT)
        self.capture.release()
        return self.frame_count

    def get_duration(self, seconds_output=True):
        self.duration = self.frame_count / self.fps
        if seconds_output:
            return self.duration
        else:
            return timedelta(seconds=self.duration)

    def sync(self, frame: int, ts):
        elapsed_time = frame / self.fps
        if type(ts) is float:
            start_time = ts - elapsed_time
        else:
            t_temp = (dateutil.parser.isoparse(ts))
            start_time = t_temp.replace(tzinfo=timezone.utc).timestamp() - elapsed_time
        self.set_start_utc(start_time)
        self.vid_summary()
        return None

    def create_pic_list_from_zip(self, i_desc_timestamps):
        intersection_desc = []
        frames = []
        prev_frame = 0
        filename_description, time_of_sd = zip(*i_desc_timestamps)

        for sd_item in range(0, len(i_desc_timestamps)):
            time_of_picture = time_of_sd[sd_item] - self.get_start_timestamp()
            if time_of_picture > 0:
                frame_of_video = time_of_picture * self.fps

                #  build up lists if not duplicate frame
                if int(frame_of_video) > int(prev_frame):
                    intersection_desc.append(filename_description[sd_item])
                    frames.append(int(frame_of_video))
                prev_frame = frame_of_video

        return intersection_desc, frames

    def extract_images(self, desc_timestamps):
        """ extract images from video based on description and timestamp zip"""
        intersection_desc, extract_frames = self.create_pic_list_from_zip(desc_timestamps)
        image_path = str(self.image_out_path) + "/" + self.video_filename + "/"
        os.makedirs(image_path, exist_ok=True)

        capture = cv2.VideoCapture(str(self.video_filepath))
        frame_count = self.get_frame_count()

        i = 0  # index for all frames to extract
        j = 0  # index for frames list to extract as image
        k = 0  # intersection string description counter

        while capture.isOpened() and len(extract_frames) > 0 and i < frame_count:

            for current_frame in tqdm(range(0, extract_frames[-1]), #tqdm(range(0, extract_frames[-1]),
                          desc="Frame Search",
                          unit=" Frames"):
                ret, frame = capture.read()
                if ret is False:
                    print("ERROR: ret is FALSE on OpenCV image")
                    break
                if i == extract_frames[j] and j <= len(extract_frames)-1:
                    cv2.imwrite(
                        image_path + str(intersection_desc[j]) + '.jpg', frame)
                    print(
                        f'PICTURE CAPTURED AT {extract_frames[j]}: {intersection_desc[j]}, Saved {j + 1} picture(s) of {len(extract_frames)}')
                    j += 1
                    k += 1
                if j == len(extract_frames):
                    print("done processing images")
                    capture.release()
                    break
                i += 1
            if i > extract_frames[-1]:
                break
        capture.release()


    def get_info(self):
        print("==================================")
        print("          VIDEO INFO              ")
        print("==================================")
        print('')
        print(f'Filename: {self.video_filename}')
        print(f'Frames per Second (FPS): {self.fps}')
        print(f'Total Number of Frames: {self.get_frame_count()}')
        print(f'Video Duration (HH:MM:SS.Z): {self.get_duration(seconds_output=False)}')
        print('')
        print("==================================")
        print("          VIDEO INFO              ")
        print("==================================")

    #  TODO: convert to start_sec, start_min=0, end_sec, end_min=0, folder="")
    def extract_frames_between(self, start_sec, end_sec, folder=""):
        """ helper function to extract frames from video during a specific time period to estimate offset
            between gpx and video

        :param start_sec:
        :param end_sec:
        :param folder:
        :return:
        """
        """ 

            ===need to know====
            time of start of gpx:
            frame # of video at a logged location on GPX file
            time of frame at location in GPX file

            Calculate:
            Big_Offset: time of frame at location in GPX file - Start of GPX file


            ============

            When is video actually available during GPX trax recording?


            seconds of video  = frame of video / fps

            Adj_Offset = time of frame at location in GPX file - seconds of video

            if adj_off < (time of frame at location in GPX file- Start time of GPX file):
                Clip GPX to shorter file that matches video.

            def video_start_utc():
                this uses gpx file and frame of gpx point to calculate when video started=

            """
        # TODO: consider another subfolder for this video filename
        image_path = './out/frames/' + self.video_filename + "/" + folder + "/"
        os.makedirs(image_path, exist_ok=True)

        print(str(self.video_filepath))

        capture = cv2.VideoCapture(str(self.video_filepath))
        #print(f'Video is Open: {self.capture.isOpened()}')

        i = 0
        start_frame = int(self.get_fps() * start_sec)
        end_frame = int(self.get_fps() * end_sec)

        while capture.isOpened():
            ret, frame = capture.read()
            if ret == False:
                break
            if start_frame <= i and i <= end_frame:
                cv2.imwrite(image_path + 'Frame' + str(i) + '.jpg', frame)
                print(f'Saved Image Frame {i} to {image_path}')
            i += 1

            if i > end_frame:
                capture.release()
                break
        capture.release()
        """
            sd_frame_num = int(time_delay * self.fps)
    
            print(f'sd_frame: {sd_frame_num}')
    
            for j in range(0, sd_frame_num)+10:
    
                ret, frame = self.video.read()
                if ret == True:
                    print(j)
                    if j == sd_frame_num:
                        cv2.imwrite("filename%d.jpg" % j, frame)
    
    
            #self.video.release()
            """

    def generate_gif_images(self, df, frame_list: list, distance):
        """ creates a folder of images to create a gif
        # /////////////*\\\\\\\\\\\\\\\
        # calculate frames needed for gif, before "/" and after frame "\" from sight Distance "*" location

        # Algo methodology
        # 1. find what frames to extract
        # 2. extract frames into folder for gif
        # 3. create gif from those frames

        :param df: dataframe of key points including speed, and descriptions of the point
        :param frame_list: list of key frame at a distance to check sight of static object
        :param distance: distance in feet before AND after of key frame to make images for
        :return: Returns a .gif filetype
        """

        # TODO: convert speed from meters/sec to feet per second
        capture = cv2.VideoCapture(self.video_filepath)
        print(f'Video is Open: {self.capture.isOpened()}')

        for i in range(0, len(frame_list)):

            folder = df.string_desc.loc[i]
            image_path = './out/frames/gif/' + folder + "/"
            os.makedirs(image_path, exist_ok=True)

            # assumes constant speed from this center point
            additional_frames = int((distance / df.spd.loc[i]) * self.fps) + 1

            print(
                f'center frame: {frame_list[i]}, additional_frame:{additional_frames}'
            )

            frame_min = 0
            frame_max = self.frame_count

            if (frame_list[i] - additional_frames) <= 0:
                frame_min = 0
            else:
                frame_min = int(frame_list[i] - additional_frames)

            if (frame_list[i] + additional_frames) >= self.frame_count:
                frame_max = int(self.frame_count)
            else:
                frame_max = int(frame_list[i] + additional_frames)

            print(
                f'min frame: {frame_min}, max frame:{frame_max}/{frame_list[i]}'
            )
            # i is df index
            j = 0  # frame index

            capture = cv2.VideoCapture(self.video_file)
            print(f'Video is Open: {self.capture.isOpened()}')

            while capture.isOpened():
                ret, frame = capture.read()
                if ret is False:
                    break
                print(f'intersection #:{i}, frame #:{j}')
                if frame_min <= j <= frame_max:
                    cv2.imwrite(
                        image_path + str(j) + ":" + df.string_desc.loc[i] +
                        '.jpg', frame)
                    print(
                        f'PICTURE CAPTURED AT FRAME:{j}: {df.string_desc.loc[i]}'
                    )
                if j > frame_max:
                    break
                else:
                    j += 1
            i += 1
        capture.release()



    @staticmethod
    def hr_min_sec(sec):
        if sec < 60:
            return f'{sec} seconds'
        elif sec < 3600:
            min = int(sec / 60)
            sec_remain = round(sec - min * 60, 2)
            return f'{min}:{sec_remain} (MM:SS.ss)'
        elif sec >= 3600:
            hr = int(sec / 3600)
            min = int(sec / 60)
            sec_remain = round(sec - min * 60, 2)
            return f'{hr}:{min}:{sec_remain} (HH:MM:SS.ss)'

    def sizeConvert(self, size):
        # size convert
        K, M, G = 1024, 1024 ** 2, 1024 ** 3
        if size >= G:
            return str(round(size / G, 2)) + ' GB'
        elif size >= M:
            return str(round(size / M, 2)) + ' MB'
        elif size >= K:
            return str(round(size / K, 2)) + ' KB'
        else:
            return str(round(size, 2)) + ' Bytes'

    def get_filesize(self):
        # get the file size
        file_byte = os.path.getsize(self.video_filepath)
        return self.sizeConvert(file_byte)

    def vid_summary(self):
        #  display values
        width = 70
        title = "VIDEO SUMMARY"
        symbol = "="

        vid_file = cv2.VideoCapture(str(self.video_filepath))
        if vid_file.isOpened():
            # get vcap property
            vid_width = vid_file.get(cv2.CAP_PROP_FRAME_WIDTH)  # float `width`
            vid_height = vid_file.get(cv2.CAP_PROP_FRAME_HEIGHT)  # float `height`

        summary = f"""
                {symbol * width}
                {" " * (int(width/2)-int(len(title)/2))}{title}
                {symbol * width}
                # Video File: {str(self.video_filepath)}
                # Video File Size: {self.get_filesize()}
                # Resolution (w x h): {vid_width} x {vid_height} ({round((vid_width * vid_height)/(1*10**6),1)}MP)
                # Frames Per Second: {self.fps}
                # Total Number of Frames: {self.frame_count}
                # Total Duration: {self.hr_min_sec(self.get_duration())}
                # Start Time: {datetime.fromtimestamp(self.start_time, tz=None)}
                # End Time:   {datetime.fromtimestamp(self.start_time + self.get_duration(), tz=None)}
                {symbol * width}
                """

        print(summary)

    @staticmethod
    def find_font_scale(label, max_width):
        font_scl = 0.1
        textsize_x, textsize_y = cv2.getTextSize(label, cv2.FONT_HERSHEY_PLAIN, font_scl, 1)[0]
        if textsize_x < max_width:
            #  scale up scale in for loop
            for scale_increment in np.arange(0, 10, 0.1):
                font_scl = scale_increment
                textsize_x, textsize_y = cv2.getTextSize(label, cv2.FONT_HERSHEY_PLAIN, font_scl, 1)[0]
                if textsize_x < max_width:
                    continue
                else:
                    font_scl = scale_increment - 0.1
                    break
        return font_scl

    def img_overlay_info_box(self, vid_filename_dir, ro_info):
        dir_as_string = str(self.image_out_path) + "/" + vid_filename_dir
        alpha = 1  # Transparency factor.
        text_font = cv2.FONT_HERSHEY_PLAIN
        font_thickness = 1

        pathlist = Path(dir_as_string).rglob('*.[0-3]-*.jpg') #  filter for images where * is wildcard
        for file in pathlist:
            filename = str(file).split("/")[-1]
            img_path = dir_as_string + "/" + filename
            img = cv2.imread(img_path)
            overlay = img.copy()

            label_img_dir = dir_as_string + "/labeled/"
            os.makedirs(label_img_dir, exist_ok=True)
            label_img_name = label_img_dir + filename

            # get img dimensions
            img_height, img_width, channels = img.shape
            rect_h = int(img_height * 0.05)
            rect_w = img_width
            rect_y = img_height-rect_h

            # build label
            sro_id = int(filename.split(".")[0])
            b_index = int((filename.rsplit(".")[1])[0:1])
            distance = 0
            ts = float(filename.split("-")[-1].replace(".jpg", ""))
            label = ro_info.intersection_frame_description(sro_id, b_index, distance, ts, desc_type="label")
            font_scale = self.find_font_scale(label, rect_w)
            textsize_x, textsize_y = cv2.getTextSize(label, text_font,  font_scale, font_thickness)[0]
            text_y = int((img_height - rect_h/2.0)+textsize_y/2.0)

            if textsize_x <= rect_w:
                text_x = int((rect_w-textsize_x)/2.0)
            else:
                text_x = 0
                label = label[0:rect_w]

            # (x,y))
            """
            0,0                                                              img_width,0
            
            
            
            0, img_height - img_height*5%             img_width, img_height - img_height*5%
            
            0,img_height                                             img_width, img_height
            """
            cv2.rectangle(overlay, pt1=(0, img_height), pt2=(rect_w, rect_y), color=(255, 255, 255), thickness=-1)
            img_new = cv2.addWeighted(overlay, alpha, img, 1-alpha, 0)
            cv2.putText(img_new, label, (text_x, text_y), text_font, font_scale, (0, 0, 0), 2)
            cv2.imwrite(label_img_name, img_new)
